/*
 Copyright(c) 2005 arton.
 
 Permission is granted for use, copying, modification, distribution,
 and distribution of modified versions of this work.
 */
using System;
using System.Collections;
using System.Diagnostics;
using System.IO;
using System.Net;
using System.Text;
using System.Text.RegularExpressions;

namespace EverMore
{
	/// <summary>
	/// y[WƂ̒̃NɑΉIuWFNgBB
	/// </summary>
	public class Node : IEnumerable, IComparable
	{
		const int BUFF_SIZE = 3200;

		internal class AsyncBuffer 
		{
			static readonly Regex CHARSET = new Regex(".+;\\s*(?i:charset)\\s*=\\s*(\\w.+)$");
			Stream respStream;
			MemoryStream memBuff;
			Encoding encoding;
			byte[] buffer;
			HttpWebRequest request;
			Node owner;

			internal AsyncBuffer(Node o, string reff)
			{
				Debug.WriteLine("start : " + o.uri.AbsoluteUri);
				owner = o;
				memBuff = new MemoryStream();
				buffer = new byte[BUFF_SIZE];
				request = (HttpWebRequest)WebRequest.Create(o.uri);
				//				request.Referer = reff;
				try
				{
					IAsyncResult r = (IAsyncResult)request.BeginGetResponse(
						new AsyncCallback(RespCallback), this);
				}
				catch (WebException e)
				{
					System.Console.WriteLine(e.StackTrace);
				}
			}

			static void RespCallback(IAsyncResult ar)
			{
				Debug.WriteLine("RespCallBack");
				AsyncBuffer a = (AsyncBuffer)ar.AsyncState;
				a.GotResponse(ar);
			}

			void GotResponse(IAsyncResult ar)
			{
				Debug.Assert(buffer != null, "bad state, no buffer");

				WebResponse resp;
				try
				{
					resp = request.EndGetResponse(ar);
				}
				catch (WebException e)
				{
					System.Console.WriteLine(e.StackTrace);
					return;
				}

				String enc = CHARSET.Match(resp.ContentType).Groups[1].Value;
				if (enc != null && enc.Length > 0)
				{
					try 
					{
						encoding = Encoding.GetEncoding(enc);
					}
					catch (ArgumentException)
					{
					}
				}
				if (encoding == null)
				{
					encoding = Encoding.UTF8;
				}

				try 
				{
					respStream = resp.GetResponseStream();
					respStream.BeginRead(buffer, 0, 
						buffer.Length, new AsyncCallback(ReadCallBack), this); 
				}
				catch (WebException e)
				{
					System.Console.WriteLine(e.StackTrace);
					resp.Close();
				}
			}

			static void ReadCallBack(IAsyncResult ar)
			{
				Debug.WriteLine("ReadCallBack");
				AsyncBuffer a = (AsyncBuffer)ar.AsyncState;
				a.GotData(ar);
			}

			void GotData(IAsyncResult ar)
			{
				Debug.Assert(encoding != null, "bad state, no encoding");

				int len;
				try
				{
					len = respStream.EndRead(ar);
				}
				catch (WebException e) 
				{
					System.Console.WriteLine(e.StackTrace);
					respStream.Close();
					return;
				}
				if (len > 0)
				{
					memBuff.Write(buffer, 0, len);
					respStream.BeginRead(buffer, 0, buffer.Length, 
						new AsyncCallback(ReadCallBack), this);
					return;
				}
				respStream.Close();
				respStream = null;

				memBuff.Flush();
				memBuff.Position = 0;
				StreamReader sr = new StreamReader(memBuff, encoding, true);
				using (sr)
				{
					owner.CreateLinks(sr.ReadToEnd());
				}
				memBuff.Close();
				memBuff = null;
			}
		}

		WebView owner;
		Uri uri;
		Hashtable links;
		ArrayList reverseLinks;
		AsyncBuffer buffer;
		int generation;

		static readonly Regex ANCHOR = new Regex("<\\s*[aA]\\s+(?i:href)\\s*=\\s*(\"|')?(http://[^\"' ]+).*?>");
		static readonly Regex[] FILTERS;
		static Node() 
		{
			FILTERS = new Regex[] {
									  new Regex("\\.(?i:yahoo)\\.[cC]"),
									  new Regex("\\.(?i:google)\\.[cC]"),
									  new Regex("\\.(?i:search)\\."),
									  new Regex("\\.(?i:amazon)\\.[cC]"),
									  new Regex("\\.(?i:goo)\\.(?i:ne)\\.(?i:jp)"),
									  new Regex("\\.(?i:excite)\\."),
									  new Regex("(?i:www\\.infoseek\\.co\\.jp)"),
									  new Regex("\\d+\\.\\d+\\.\\d+\\.\\d+"),
								};
		}

		public Node(WebView initOwner, Uri initUri, int initgen)
		{
			Debug.WriteLine("create Node:" + initUri.AbsoluteUri);
			generation = initgen;
			links = new Hashtable();
			reverseLinks = new ArrayList();
			owner = initOwner;
			uri = initUri;
		}

		public void StartRead(string reff)
		{
			if (!Filtered(uri) && generation < owner.Depth)
			{
				buffer = new AsyncBuffer(this, reff);
			}
		}

		bool Filtered(Uri uri)
		{
			if (generation == 0)
			{
				return false;
			}
			for (int i = 0; i < FILTERS.Length; i++)
			{
				if (FILTERS[i].IsMatch(uri.Host))
				{
					return true;
				}
			}
			return false;
		}

		public Uri Uri
		{
			get
			{
				return uri;
			}
		}
		public int Count
		{
			get
			{
				Debug.Assert(links != null, "not initialized");
				return links.Count;
			}
		}

		public int Generation
		{
			get
			{
				return generation;
			}
		}

		public IEnumerator GetEnumerator()
		{
			return links.GetEnumerator();
		}

		public IEnumerable GetReveseLinks()
		{
			return reverseLinks;
		}

		public IEnumerable GetLinks()
		{
			return links.Values;
		}

		public string AbsoluteUri
		{
			get
			{
				return uri.AbsoluteUri;
			}
		}

		void AddReverse(Node n)
		{
			if (n.Equals(this))
			{
				return;
			}
			lock (reverseLinks)
			{
				if (!reverseLinks.Contains(n))
				{
					reverseLinks.Add(n);
				}
			}
		}

		private void CreateLinks(String html)
		{
			MatchCollection mc = ANCHOR.Matches(html);
			foreach (Match m in mc)
			{
				string url = m.Groups[2].Value;
				Debug.WriteLine("url=" + url);
				Uri uri;
				try
				{
					uri = new Uri(url);
				}
				catch (UriFormatException)
				{
					continue;
				}
				if (!uri.Scheme.Equals(Uri.UriSchemeHttp))
				{
					continue;
				}
				if (links.ContainsKey(uri.AbsoluteUri))
				{
					continue;
				}
				Node n = owner.CreateNode(uri, AbsoluteUri, generation + 1);
				if (n == null || n.Equals(this))
				{
					continue;
				}
				links.Add(uri.AbsoluteUri, n);
				n.AddReverse(this);
			}
		}
		#region IComparable o

		public int CompareTo(object obj)
		{
			// TODO:  Node.CompareTo ǉ܂B
			if (obj == null || !typeof(Node).Equals(obj.GetType()))
			{
				throw new ArgumentException("bad type of " + obj);
			}
			return AbsoluteUri.CompareTo(((Node)obj).AbsoluteUri);
		}

		#endregion
	}
}
